/* vi: set sw=4 ts=4: */
/* ARM ELF shared library loader suppport
 *
 * Copyright (C) 2001-2004 Erik Andersen
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the above contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

static const char * const _dl_reltypes_tab[] =
{
  [0]	"R_AARCH64_NONE",	"R_AARCH64_P32_ABS32",
  [180]	"R_AARCH64_P32_COPY", "R_AARCH64_P32_GLOB_DAT", "R_AARCH64_P32_JUMP_SLOT",
  [183]	"R_AARCH64_P32_RELATIVE", "R_AARCH64_P32_TLS_DTPMOD", "R_AARCH64_P32_TLS_DTPREL",
  [186]	"R_AARCH64_P32_TLS_TPREL", "R_AARCH64_P32_TLSDESC", "R_AARCH64_P32_IRELATIVE",

  [257]	"R_AARCH64_ABS64", "R_AARCH64_ABS32", "R_AARCH64_ABS16", "R_AARCH64_PREL64",
  [261]	"R_AARCH64_PREL32", "R_AARCH64_PREL16", "R_AARCH64_MOVW_UABS_G0", "R_AARCH64_MOVW_UABS_G0_NC",
  [265]	"R_AARCH64_MOVW_UABS_G1", "R_AARCH64_MOVW_UABS_G1_NC", "R_AARCH64_MOVW_UABS_G2",
  [268]	"R_AARCH64_MOVW_UABS_G2_NC", "R_AARCH64_MOVW_UABS_G3", "R_AARCH64_MOVW_SABS_G0",
  [271]	"R_AARCH64_MOVW_SABS_G1", "R_AARCH64_MOVW_SABS_G2", "R_AARCH64_LD_PREL_LO19",
  [274]	"R_AARCH64_ADR_PREL_LO21", "R_AARCH64_ADR_PREL_PG_HI21", "R_AARCH64_ADR_PREL_PG_HI21_NC",
  [277]	"R_AARCH64_ADD_ABS_LO12_NC", "R_AARCH64_LDST8_ABS_LO12_NC", "R_AARCH64_TSTBR14",
  [280]	"R_AARCH64_CONDBR19",
  [282]	"R_AARCH64_JUMP26", "R_AARCH64_CALL26", "R_AARCH64_LDST16_ABS_LO12_NC",
  [285]	"R_AARCH64_LDST32_ABS_LO12_NC", "R_AARCH64_LDST64_ABS_LO12_NC", "R_AARCH64_MOVW_PREL_G0",
  [288]	"R_AARCH64_MOVW_PREL_G0_NC", "R_AARCH64_MOVW_PREL_G1", "R_AARCH64_MOVW_PREL_G1_NC",
  [291]	"R_AARCH64_MOVW_PREL_G2", "R_AARCH64_MOVW_PREL_G2_NC", "R_AARCH64_MOVW_PREL_G3",
  [299]	"R_AARCH64_LDST128_ABS_LO12_NC", "R_AARCH64_MOVW_GOTOFF_G0", "R_AARCH64_MOVW_GOTOFF_G0_NC",
  [302]	"R_AARCH64_MOVW_GOTOFF_G1", "R_AARCH64_MOVW_GOTOFF_G1_NC", "R_AARCH64_MOVW_GOTOFF_G2",
  [305]	"R_AARCH64_MOVW_GOTOFF_G2_NC", "R_AARCH64_MOVW_GOTOFF_G3", "R_AARCH64_GOTREL64",
  [308]	"R_AARCH64_GOTREL32", "R_AARCH64_GOT_LD_PREL19", "R_AARCH64_LD64_GOTOFF_LO15",
  [311]	"R_AARCH64_ADR_GOT_PAGE", "R_AARCH64_LD64_GOT_LO12_NC", "R_AARCH64_LD64_GOTPAGE_LO15",

  [512]	"R_AARCH64_TLSGD_ADR_PREL21", "R_AARCH64_TLSGD_ADR_PAGE21", "R_AARCH64_TLSGD_ADD_LO12_NC",
  [515]	"R_AARCH64_TLSGD_MOVW_G1", "R_AARCH64_TLSGD_MOVW_G0_NC", "R_AARCH64_TLSLD_ADR_PREL21",
  [518]	"R_AARCH64_TLSLD_ADR_PAGE21", "R_AARCH64_TLSLD_ADD_LO12_NC", "R_AARCH64_TLSLD_MOVW_G1",
  [521]	"R_AARCH64_TLSLD_MOVW_G0_NC", "R_AARCH64_TLSLD_LD_PREL19", "R_AARCH64_TLSLD_MOVW_DTPREL_G2",
  [524]	"R_AARCH64_TLSLD_MOVW_DTPREL_G1", "R_AARCH64_TLSLD_MOVW_DTPREL_G1_NC",
  [526]	"R_AARCH64_TLSLD_MOVW_DTPREL_G0", "R_AARCH64_TLSLD_MOVW_DTPREL_G0_NC",
  [528]	"R_AARCH64_TLSLD_ADD_DTPREL_HI12", "R_AARCH64_TLSLD_ADD_DTPREL_LO12",
  [530]	"R_AARCH64_TLSLD_ADD_DTPREL_LO12_NC", "R_AARCH64_TLSLD_LDST8_DTPREL_LO12",
  [532]	"R_AARCH64_TLSLD_LDST8_DTPREL_LO12_NC", "R_AARCH64_TLSLD_LDST16_DTPREL_LO12",
  [534]	"R_AARCH64_TLSLD_LDST16_DTPREL_LO12_NC", "R_AARCH64_TLSLD_LDST32_DTPREL_LO12",
  [536]	"R_AARCH64_TLSLD_LDST32_DTPREL_LO12_NC", "R_AARCH64_TLSLD_LDST64_DTPREL_LO12",
  [538]	"R_AARCH64_TLSLD_LDST64_DTPREL_LO12_NC", "R_AARCH64_TLSIE_MOVW_GOTTPREL_G1",
  [540]	"R_AARCH64_TLSIE_MOVW_GOTTPREL_G0_NC", "R_AARCH64_TLSIE_ADR_GOTTPREL_PAGE21",
  [542]	"R_AARCH64_TLSIE_LD64_GOTTPREL_LO12_NC", "R_AARCH64_TLSIE_LD_GOTTPREL_PREL19",
  [544]	"R_AARCH64_TLSLE_MOVW_TPREL_G2", "R_AARCH64_TLSLE_MOVW_TPREL_G1",
  [546]	"R_AARCH64_TLSLE_MOVW_TPREL_G1_NC", "R_AARCH64_TLSLE_MOVW_TPREL_G0",
  [548]	"R_AARCH64_TLSLE_MOVW_TPREL_G0_NC", "R_AARCH64_TLSLE_ADD_TPREL_HI12",
  [550]	"R_AARCH64_TLSLE_ADD_TPREL_LO12", "R_AARCH64_TLSLE_ADD_TPREL_LO12_NC",
  [552]	"R_AARCH64_TLSLE_LDST8_TPREL_LO12", "R_AARCH64_TLSLE_LDST8_TPREL_LO12_NC",
  [554]	"R_AARCH64_TLSLE_LDST16_TPREL_LO12", "R_AARCH64_TLSLE_LDST16_TPREL_LO12_NC",
  [556]	"R_AARCH64_TLSLE_LDST32_TPREL_LO12", "R_AARCH64_TLSLE_LDST32_TPREL_LO12_NC",
  [558]	"R_AARCH64_TLSLE_LDST64_TPREL_LO12", "R_AARCH64_TLSLE_LDST64_TPREL_LO12_NC",
  [560]	"R_AARCH64_TLSDESC_LD_PREL19", "R_AARCH64_TLSDESC_ADR_PREL21", "R_AARCH64_TLSDESC_ADR_PAGE21",
  [563]	"R_AARCH64_TLSDESC_LD64_LO12", "R_AARCH64_TLSDESC_ADD_LO12", "R_AARCH64_TLSDESC_OFF_G1",
  [566]	"R_AARCH64_TLSDESC_OFF_G0_NC", "R_AARCH64_TLSDESC_LDR", "R_AARCH64_TLSDESC_ADD",
  [569]	"R_AARCH64_TLSDESC_CALL", "R_AARCH64_TLSLE_LDST128_TPREL_LO12",
  [571]	"R_AARCH64_TLSLE_LDST128_TPREL_LO12_NC", "R_AARCH64_TLSLD_LDST128_DTPREL_LO12",
  [573]	"R_AARCH64_TLSLD_LDST128_DTPREL_LO12_NC",
  [1024]	"R_AARCH64_COPY", "R_AARCH64_GLOB_DAT", "R_AARCH64_JUMP_SLOT", "R_AARCH64_RELATIVE",
  [1028]	"R_AARCH64_TLS_DTPMOD", "R_AARCH64_TLS_DTPREL", "R_AARCH64_TLS_TPREL",
  [1031]	"R_AARCH64_TLSDESC", "R_AARCH64_IRELATIVE",
};
